"""
Fetch current lightcurve from Flows API.
"""

import os.path
import tempfile
from astropy.table import Table
from tendrils.utils import get_api_token, get_request, URLS
from typing import Union


def get_lightcurve(target: Union[int, str]) -> Table:
    """
    Retrieve lightcurve from Flows server.

    Parameters:
        target (int): Target to download lightcurve for.

    Returns:
        :class:`astropy.table.Table`: Table containing lightcurve.

    TODO:
        - Enable caching of files.
    """
    token = get_api_token()
    # Send query to the Flows API:
    params = {'target': target}
    r = get_request(URLS.lightcurves_url, token=token, params=params)
    text = r.text.replace('str,', 'string,')  # only needed for astropy >=4.3 and < 5.1
    # Create temp directory and save the file into there,
    # then open the file as a Table:
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpfile = os.path.join(tmpdir, 'table.ecsv')
        with open(tmpfile, 'w') as fid:
            fid.write(text)

        tab = Table.read(tmpfile, format='ascii.ecsv')

    return tab



"""
(Cosmic) Tendrils is the generic API for (cosmic) FLOWS project and pipeline.
"""
from .version import version as __version__
from .version import version_info as __version_info__
import tendrils.api
import tendrils.utils



"""
For querying site information
"""

from functools import lru_cache
import astropy.units as u
from astropy.coordinates import EarthLocation
from tendrils.utils import get_api_token, get_request, URLS
from typing import Union


@lru_cache(maxsize=10)
def get_site(siteid: Union[int, str]) -> dict:
    """
    Get site information (defined per each instrument).
    Args:
        siteid: Optional[int, str] site id number.

    Returns: json of site (instrument) parameters.

    """
    token = get_api_token()
    r = get_request(URLS.sites_url, token=token, params={'siteid': siteid})
    jsn = r.json()

    # Special derived objects:
    jsn['EarthLocation'] = EarthLocation(lat=jsn['latitude'] * u.deg, lon=jsn['longitude'] * u.deg,
                                         height=jsn['elevation'] * u.m)
    return jsn


@lru_cache(maxsize=1)
def get_all_sites() -> dict:
    """
    Get info for all sites registered in FLOWS
    Returns: json of site (instrument) parameters.

    """
    token = get_api_token()
    r = get_request(URLS.sites_url, token=token)
    jsn = r.json()

    # Special derived objects:
    for site in jsn:
        site['EarthLocation'] = EarthLocation(lat=site['latitude'] * u.deg, lon=site['longitude'] * u.deg,
                                              height=site['elevation'] * u.m)
    return jsn



"""
Functions for dealing with configuration of Tendrils API using .ini files or secrets.
"""

import configparser
import os.path
from functools import lru_cache
from typing import Optional
from warnings import warn


@lru_cache(maxsize=1)
def load_config(filename: str = 'config.ini') -> configparser.ConfigParser:
    """
    Load configuration file. Defaults to searching for config.ini but a custom name can be given.

    Returns:
        ``configparser.ConfigParser``: Configuration file.

    .. codeauthor:: Rasmus Handberg <rasmush@phys.au.dk>
    .. codeauthor:: Emir Karamehmetoglu <emir.k@phys.au.dk>
    """

    config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    if not os.path.isfile(config_file):
        raise FileNotFoundError(f"{filename} file not found")

    config = configparser.ConfigParser()
    config.read(config_file)

    if config.get('api','token') == 'None':
        warn("Token could not be found! Try using 'set_api_token', 'create_config', or 'copy_from_other_config'")
    return config


def clear_config_cache():
    """
    Clear the config cache.
    """
    load_config.cache_clear()


def set_api_token(token: Optional[str] = None, filename: Optional[str] = 'config.ini', overwrite: bool = True) -> None:
    """
    For updating FLOWS API token.
    Args:
        token (str, optional): FLOWS api token
        filename (str): filename of config file, defaults to `config.ini`.
        overwrite (bool): Whether to overwrite the config if the field is not "None".
        True by default. Set to false if you don't want to overwrite if field is not "None".
    """
    config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    config = configparser.ConfigParser()
    config.read(config_file)

    if config.get('api', 'token') == "None" or overwrite:
        if token is None:
            token =  input('Enter API token to be saved into config:')
        config.set('api','token', token)

    with open(config_file, 'w') as cfgfile:
        config.write(cfgfile)

    clear_config_cache()


def update_api_token() -> str:
    """
    Update API token in config file via explicit query.
    """
    token = str(input('Tendrils needs your FLOWS API token. Please enter it:'))
    # We remove spaces from the token to avoid errors.
    token = token.strip().replace(' ', '')
    set_api_token(token, overwrite=True)
    print(f'Token with value {token} set and saved.\nIf there is a problem, please run Tendrils again.')
    return token


def get_api_token() -> str:
    """
    Get api token from config file or raise.
    Returns: str = token as a string

    """
    # Get API token from config file:
    config = load_config()
    token = config.get('api', 'token', fallback=None)

    if token is None:
        raise RuntimeError("No API token has been defined")

    if token.lower() in ['none', 'test', '', 'bad', 'bad_token', 'badtoken']:
        token = update_api_token()

    return token


def set_photometry_folders(output: Optional[str] = None, archive_local: Optional[str] = None,
                           filename: str = 'config.ini', overwrite:bool = True) -> None:
    """
    For updating FLOWS photometry output and local archive folders.
    Args:
        output (str, optional): path to desired photometry output directory.
        archive_local (str, optional): path to desired local archive directory.
        filename (str): default = 'config.ini'. Filename for config file
        overwrite (bool): Whether to overwrite existing field if field is not "None".
    """

    config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    config = configparser.ConfigParser()
    config.read(config_file)

    if config.get('photometry', 'output') == "None" or overwrite:
        if output is None:
            config['photometry']['output'] = input('Enter photometry output directory absolute path:')
        else:
            config['photometry']['output'] = output

    if config.get('photometry', 'archive_local') == "None" or overwrite:
        if archive_local is None:
            config['photometry']['archive_local'] = input('Enter local data archive directory absolute path:')
        else:
            config['photometry']['archive_local'] = archive_local

    with open(config_file, 'w') as cfgfile:
        config.write(cfgfile)

    clear_config_cache()


def set_tns_token(api_key: Optional[str] = None, filename: str = 'config.ini', overwrite:bool = False) -> None:
    config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    config = configparser.ConfigParser()
    config.read(config_file)

    if config.get('TNS', 'api_key') == "None" or overwrite:
        if api_key is None:
            config['TNS']['api_key'] = input('Enter photometry output directory absolute path:')
        else:
            config['TNS']['api_key'] = api_key

    with open(config_file, 'w') as cfgfile:
        config.write(cfgfile)

    clear_config_cache()


def create_config(tns:bool = False):
    set_api_token(overwrite=True)
    set_photometry_folders(overwrite=True)
    if tns:
        set_tns_token(overwrite=True)
    clear_config_cache()


def copy_from_other_config(filepath: str, filename: str = 'config.ini'):
    config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), filename)
    config_sourcefile = os.path.expanduser(filepath)

    config_source = configparser.ConfigParser()
    config_dest= configparser.ConfigParser()

    config_source.read(config_sourcefile)
    config_dest.read(config_file)

    config_dest['api'] = config_source['api']
    config_dest['photometry'] = config_source['photometry']
    config_dest['TNS'] = config_source['TNS']

    with open(config_file, 'w') as cfgfile:
        config_dest.write(cfgfile)
    clear_config_cache()



"""
Upload photometry results to Flows server.
"""

import glob
import logging
import os
import shutil
import zipfile
import tempfile
from typing import Union, Optional
from configparser import ConfigParser
from astropy.table import Table
from tqdm import tqdm

from tendrils.api import get_datafile
from tendrils.utils import get_api_token, get_request, load_config, get_filehash, post_request, URLS


def get_photcache(config: Optional[ConfigParser] = None) -> tuple[str, Optional[tempfile.TemporaryDirectory]]:
    """
    Get photometry cache from config or create a temporary one
    Args:
        config: ConfigParser instance with 'api' and 'photometry_cache' defined.

    Returns: tuple[str, str]: tuple (photcache path, reference to tmpdir).
     If photometry_cache was supplied via config, the latter is None.
    """
    # Use config if given:
    if config is not None:
        photcache = config.get('api', 'photometry_cache', fallback=None)
    else:
        photcache = None

    # Try config location if exists, default to tempdir if not valid:
    tmpdir = None
    if photcache is not None:
        photcache = os.path.abspath(photcache)
        if not os.path.isdir(photcache):
            raise FileNotFoundError(f"Photometry cache directory does not exist: {photcache}")
    else:
        tmpdir = tempfile.TemporaryDirectory(prefix='flows-api-get_photometry-')
        photcache = tmpdir.name

    return photcache, tmpdir


def create_photdir(config: ConfigParser,
                   target_name: str,
                   fileid: Union[int, str]) -> str:
    """
    Get photdir for given fileid and target_name
    Args:
        fileid (int, str): fileid
        target_name (str): target_name
        config (configparser.ConfigParser): current config instance
    Returns: photdir: absolute path to photdir.

    """
    fileid_str = f'{int(fileid):05d}'
    photdir_root = config.get('photometry', 'output', fallback='.')

    # Find the photometry output directory for this fileid:
    photdir = os.path.join(photdir_root, target_name, fileid_str)
    if not os.path.isdir(photdir):
        # Do a last check, to ensure that we have not just added the wrong number of zeros
        # to the directory name:
        found_photdir = []
        for d in os.listdir(os.path.join(photdir_root, target_name)):
            if d.isdigit() and int(d) == fileid and os.path.isdir(d):
                found_photdir.append(os.path.join(photdir_root, target_name, d))
        # If we only found one, use it, otherwise throw an exception:
        if len(found_photdir) == 1:
            photdir = found_photdir[0]
        elif len(found_photdir) > 1:
            raise RuntimeError(f"Several photometry output found for fileid={fileid}. \
                    You need to do a cleanup of the photometry output directories.")
        else:
            raise FileNotFoundError(photdir)

    photdir = os.path.abspath(photdir)
    return photdir


def make_zip(files: list, fileid: Union[int, str], current_dir: Union[tempfile.TemporaryDirectory, str]):
    # Logging and tqdm
    logger = logging.getLogger(__name__)
    tqdm_settings = {'disable': None if logger.isEnabledFor(logging.INFO) else True}

    # Create ZIP-file within the temp directory:
    fpath_zip = os.path.join(current_dir, f'{int(fileid):05d}.zip')

    # Create ZIP file with all the files:
    with zipfile.ZipFile(fpath_zip, 'w', allowZip64=True) as z:
        for f in tqdm(files, desc=f'Zipping {int(fileid):d}', **tqdm_settings):
            logger.debug('Zipping %s', f)
            z.write(f, os.path.basename(f))

    # Change the name of the uploaded file to contain the file hash:
    fhash = get_filehash(fpath_zip)
    fname_zip = f'{int(fileid):05d}-{fhash:s}.zip'

    return fhash, fname_zip


def get_photometry(photid: int) -> Table:
    """
    Retrieve lightcurve from Flows server.

    Please note that it can significantly speed up repeated calls to this function
    to specify a cache directory in the config-file under api -> photometry_cache.
    This will download the files only once and store them in this local cache for
    use in subsequent calls.

    Parameters:
        photid (int): Fileid for the photometry file.

    Returns:
        :class:`astropy.table.Table`: Table containing photometry.
    """
    token = get_api_token()
    config = load_config()
    photcache, tmpdir = get_photcache(config)

    # Construct path to the photometry file in the cache:
    photfile = os.path.join(photcache, f'photometry-{photid:d}.ecsv')

    if not os.path.isfile(photfile):
        # Send query to the Flows API:
        params = {'fileid': photid}
        r = get_request(URLS.photometry_url, token=token, params=params)

        # Create temporary directory and save the file into there,
        # then open the file as a Table:
        with open(photfile, 'w') as fid:
            fid.write(r.text)

    # Read the photometry file:
    tab = Table.read(photfile, format='ascii.ecsv')

    # Explicitly cleanup the temporary directory if it was created:
    if tmpdir:
        tmpdir.cleanup()

    return tab


def upload_photometry(fileid: Union[int, str], delete_completed: bool = False) -> None:
    """
    Upload photometry results to Flows server.

    This will make the uploaded photometry the active/newest/best photometry and
    be used in plots and shown on the website.

    Parameters:
        fileid: [int, str]: File ID of photometry to upload to server.
        delete_completed: bool, optional: Delete the photometry from the local
            working directory if the upload was successful. Default=False.
    """

    logger = logging.getLogger(__name__)

    # Use API to get the datafile information:
    datafile = get_datafile(fileid)
    token = get_api_token()
    config = load_config()
    photdir = create_photdir(config, datafile['target_name'], fileid)

    # Make sure required files are actually there:
    files_existing = os.listdir(photdir)
    if 'photometry.ecsv' not in files_existing:
        raise FileNotFoundError(os.path.join(photdir, 'photometry.ecsv'))
    if 'photometry.log' not in files_existing:
        raise FileNotFoundError(os.path.join(photdir, 'photometry.log'))

    # Create list of files to be uploaded:
    files = [os.path.join(photdir, 'photometry.ecsv'), os.path.join(photdir, 'photometry.log')]
    files += glob.glob(os.path.join(photdir, '*.png'))

    # Create ZIP file:
    with tempfile.TemporaryDirectory(prefix='flows-upload-') as tmpdir:
        # Create ZIP-file within the temp directory:
        fhash, fname_zip = make_zip(files, fileid, tmpdir)
        fpath_zip = os.path.join(tmpdir, f'{fileid:05d}.zip')

        # Send file to the API:
        logger.info("Uploading to server...")
        with open(fpath_zip, 'rb') as fid:
            r = post_request(URLS.photometry_upload_url, token=token, params={'fileid': fileid},
                             files={'file': (fname_zip, fid, 'application/zip')})

    # Check the returned data from the API:
    if r.text.strip() != 'OK':
        logger.error(r.text)
        raise RuntimeError("An error occurred while uploading photometry: " + r.text)

    # If we have made it this far, the upload must have been a success:
    if delete_completed:
        if set([os.path.basename(f) for f in files]) == set(os.listdir(photdir)):
            logger.info("Deleting photometry from workdir: '%s'", photdir)
            shutil.rmtree(photdir, ignore_errors=True)
        else:
            logger.warning("Not deleting photometry from workdir: '%s'", photdir)



"""
TNS API FUNCTIONS
Pre-provided helper functions for the TNS API, type annotations added
Obtained from https://wis-tns.weizmann.ac.il/content/tns-getting-started
@TODO: Refactor and change to TNS class
"""
from __future__ import annotations
import logging
from astropy.table import Table
import astropy.units as u
from astropy.coordinates import SkyCoord
import requests
import json
import datetime
from tendrils.utils.config import load_config
from typing import Optional, Union

url_tns_api = 'https://www.wis-tns.org/api/get'
url_tns_search = 'https://www.wis-tns.org/search'
DateType = Union[datetime.datetime, str]


class TNSConfigError(RuntimeError):
    pass


def load_tns_config() -> dict[str, str]:
    logger = logging.getLogger(__name__)

    config = load_config()
    api_key = config.get('TNS', 'api_key', fallback=None)
    if api_key is None:
        raise TNSConfigError("No TNS API-KEY has been defined in config")

    tns_bot_id = config.getint('TNS', 'bot_id', fallback=93222)
    tns_bot_name = config.get('TNS', 'bot_name', fallback='AUFLOWS_BOT')
    tns_user_id = config.getint('TNS', 'user_id', fallback=None)
    tns_user_name = config.get('TNS', 'user_name', fallback=None)

    if tns_user_id and tns_user_name:
        logger.debug('Using TNS credentials: user=%s', tns_user_name)
        user_agent = 'tns_marker{"tns_id":' + str(tns_user_id) + ',"type":"user","name":"' + tns_user_name + '"}'
    elif tns_bot_id and tns_bot_name:
        logger.debug('Using TNS credentials: bot=%s', tns_bot_name)
        user_agent = 'tns_marker{"tns_id":' + str(tns_bot_id) + ',"type":"bot","name":"' + tns_bot_name + '"}'
    else:
        raise TNSConfigError("No TNS bot_id or bot_name has been defined in config")

    return {'api-key': api_key, 'user-agent': user_agent}


def tns_search(coord: Optional[SkyCoord] = None, radius: u.Quantity = 3 * u.arcsec, objname: Optional[str] = None,
               internal_name: Optional[str] = None) -> Optional[dict]:
    """
    Cone-search TNS for object near coordinate.

    Parameters:
        coord (:class:`astropy.coordinates.SkyCoord`): Central coordinate to search around.
        radius (Angle, optional): Radius to search around ``coord``.
        objname (str, optional): Search on object name.
        internal_name (str, optional): Search on internal name.

    Returns:
        dict: Dictionary with TSN response.
    """

    # API key for Bot
    tnsconf = _load_tns_config()

    # change json_list to json format
    json_file = {'radius': radius.to('arcsec').value, 'units': 'arcsec', 'objname': objname,
                 'internal_name': internal_name}
    if coord:
        json_file['ra'] = coord.icrs.ra.deg
        json_file['dec'] = coord.icrs.dec.deg

    # construct the list of (key,value) pairs
    headers = {'user-agent': tnsconf['user-agent']}
    search_data = [('api_key', (None, tnsconf['api-key'])), ('data', (None, json.dumps(json_file)))]

    # search obj using request module
    res = requests.post(url_tns_api + '/search', files=search_data, headers=headers)
    res.raise_for_status()
    parsed = res.json()
    data = parsed['data']

    if 'reply' in data:
        return data['reply']
    return None


def tns_get_obj(name:str) -> Optional[dict]:
    """
    Search TNS for object by name.

    Parameters:
        name (str): Object name to search for.

    Returns:
        dict: Dictionary with TSN response.
    """

    # API key for Bot
    tnsconf = _load_tns_config()

    # construct the list of (key,value) pairs
    headers = {'user-agent': tnsconf['user-agent']}
    params = {'objname': name, 'photometry': '0', 'spectra': '0'}
    get_data = [('api_key', (None, tnsconf['api-key'])), ('data', (None, json.dumps(params)))]

    # get obj using request module
    res = requests.post(url_tns_api + '/object', files=get_data, headers=headers)
    res.raise_for_status()
    parsed = res.json()
    data = parsed['data']

    if 'reply' in data:
        reply = data['reply']
        if not reply:
            return None
        if 'objname' not in reply:  # Bit of a cheat, but it is simple and works
            return None

        reply['internal_names'] = [name.strip() for name in reply['internal_names'].split(',') if name.strip()]
        return reply
    return None


def tns_getnames(months: Optional[int] = None, date_begin: Optional[DateType] = None,
                 date_end: Optional[DateType] = None, zmin: Optional[float] = None,
                 zmax: Optional[float] = None, objtype: tuple[int] = (3, 104)) -> list[str]:
    """
    Get SN names from TNS.

    Parameters:
        months (int, optional): Only return objects reported within the last X months.
        date_begin (date, optional): Discovery date begin.
        date_end (date, optional): Discovery date end.
        zmin (float, optional): Minimum redshift.
        zmax (float, optional): Maximum redshift.
        objtype (list, optional): Constraint object type.
            Default is to query for
            - 3: SN Ia
            - 104: SN Ia-91T-like

    Returns:
        list: List of names fulfilling search criteria.
    """

    logger = logging.getLogger(__name__)

    # Change formats of input to be ready for query:
    if isinstance(date_begin, datetime.datetime):
        date_begin = date_begin.date()
    elif isinstance(date_begin, str):
        date_begin = datetime.datetime.strptime(date_begin, '%Y-%m-%d').date()

    if isinstance(date_end, datetime.datetime):
        date_end = date_end.date()
    elif isinstance(date_end, str):
        date_end = datetime.datetime.strptime(date_end, '%Y-%m-%d').date()

    if isinstance(objtype, (list, tuple)):
        objtype = ','.join([str(o) for o in objtype])

    # Do some sanity checks:
    if date_end < date_begin:
        raise ValueError("Dates are in the wrong order.")

    date_now = datetime.datetime.now(datetime.timezone.utc).date()
    if months is not None and date_end is not None and date_end < date_now - datetime.timedelta(days=months * 30):
        logger.warning('Months limit restricts days_begin, consider increasing limit_months.')

    # API key for Bot
    tnsconf = _load_tns_config()

    # Parameters for query:
    params = {'discovered_period_value': months,  # Reported Within The Last
              'discovered_period_units': 'months', 'unclassified_at': 0,  # Limit to unclasssified ATs
              'classified_sne': 1,  # Limit to classified SNe
              'include_frb': 0,  # Include FRBs
              # 'name': ,
              'name_like': 0, 'isTNS_AT': 'all', 'public': 'all',  # 'ra':
              # 'decl':
              # 'radius':
              # 'coords_unit': 'arcsec',
              'reporting_groupid[]': 'null', 'groupid[]': 'null', 'classifier_groupid[]': 'null', 'objtype[]': objtype,
              'at_type[]': 'null', 'date_start[date]': date_begin.isoformat(), 'date_end[date]': date_end.isoformat(),
              # 'discovery_mag_min':
              # 'discovery_mag_max':
              # 'internal_name':
              # 'discoverer':
              # 'classifier':
              # 'spectra_count':
              'redshift_min': zmin, 'redshift_max': zmax,  # 'hostname':
              # 'ext_catid':
              # 'ra_range_min':
              # 'ra_range_max':
              # 'decl_range_min':
              # 'decl_range_max':
              'discovery_instrument[]': 'null', 'classification_instrument[]': 'null', 'associated_groups[]': 'null',
              # 'at_rep_remarks':
              # 'class_rep_remarks':
              # 'frb_repeat': 'all'
              # 'frb_repeater_of_objid':
              'frb_measured_redshift': 0,  # 'frb_dm_range_min':
              # 'frb_dm_range_max':
              # 'frb_rm_range_min':
              # 'frb_rm_range_max':
              # 'frb_snr_range_min':
              # 'frb_snr_range_max':
              # 'frb_flux_range_min':
              # 'frb_flux_range_max':
              'num_page': 500, 'display[redshift]': 0, 'display[hostname]': 0, 'display[host_redshift]': 0,
              'display[source_group_name]': 0, 'display[classifying_source_group_name]': 0,
              'display[discovering_instrument_name]': 0, 'display[classifing_instrument_name]': 0,
              'display[programs_name]': 0, 'display[internal_name]': 0, 'display[isTNS_AT]': 0, 'display[public]': 0,
              'display[end_pop_period]': 0, 'display[spectra_count]': 0, 'display[discoverymag]': 0,
              'display[discmagfilter]': 0, 'display[discoverydate]': 0, 'display[discoverer]': 0, 'display[remarks]': 0,
              'display[sources]': 0, 'display[bibcode]': 0, 'display[ext_catalogs]': 0, 'format': 'csv'}

    # Query TNS for names:
    headers = {'user-agent': tnsconf['user-agent']}
    con = requests.get(url_tns_search, params=params, headers=headers)
    con.raise_for_status()

    # Parse the CSV table:
    # Ensure that there is a newline in table string.
    # AstroPy uses this to distinguish file-paths from pure-string inputs:
    text = str(con.text) + "\n"
    tab = Table.read(text, format='ascii.csv', guess=False, delimiter=',', quotechar='"', header_start=0, data_start=1)

    # Pull out the names only if they begin with "SN":
    names_list = [name.replace(' ', '') for name in tab['Name'] if name.startswith('SN')]
    names_list = sorted(names_list)

    return names_list



"""
Get FLOWS specific catalogs.
"""

from functools import lru_cache
import astropy.units as u
from astropy.table import Table
from astropy.time import Time
from typing import Union, Optional
from tendrils.utils import get_api_token, get_request, URLS
from enum import Enum


class OutputFormat(Enum):
    table = 'table'
    dictionary = 'dict'
    json = 'json'


def create_catalog_table(jsn: dict) -> dict:
    dict_tables: dict[str, Table] = {}

    tab = Table(names=(
        'targetid', 'target_name', 'target_status', 'ra', 'decl', 'redshift', 'redshift_error', 'discovery_mag',
        'catalog_downloaded', 'pointing_model_created', 'inserted', 'discovery_date', 'project', 'host_galaxy',
        'ztf_id', 'sntype'), dtype=(
        'int32', 'str', 'str', 'float64', 'float64', 'float32', 'float32', 'float32', 'bool', 'bool', 'object',
        'object', 'str', 'str', 'str', 'str'), rows=[jsn['target']])

    tab['ra'].description = 'Right ascension'
    tab['ra'].unit = u.deg
    tab['decl'].description = 'Declination'
    tab['decl'].unit = u.deg
    dict_tables['target'] = tab

    for table_name in ('references', 'avoid'):
        tab = Table(names=(
            'starid', 'ra', 'decl', 'pm_ra', 'pm_dec', 'gaia_mag', 'gaia_bp_mag', 'gaia_rp_mag', 'gaia_variability',
            'B_mag', 'V_mag', 'H_mag', 'J_mag', 'K_mag', 'u_mag', 'g_mag', 'r_mag', 'i_mag', 'z_mag', 'distance'),
            dtype=('int64', 'float64', 'float64', 'float32', 'float32', 'float32', 'float32', 'float32', 'int32',
                   'float32', 'float32', 'float32', 'float32', 'float32', 'float32', 'float32', 'float32', 'float32',
                   'float32', 'float64'), rows=jsn[table_name])

        tab['starid'].description = 'Unique identifier in REFCAT2 catalog'
        tab['ra'].description = 'Right ascension'
        tab['ra'].unit = u.deg
        tab['decl'].description = 'Declination'
        tab['decl'].unit = u.deg
        tab['pm_ra'].description = 'Proper motion in right ascension'
        tab['pm_ra'].unit = u.mas / u.yr
        tab['pm_dec'].description = 'Proper motion in declination'
        tab['pm_dec'].unit = u.mas / u.yr
        tab['distance'].description = 'Distance from object to target'
        tab['distance'].unit = u.deg

        tab['gaia_mag'].description = 'Gaia G magnitude'
        tab['gaia_bp_mag'].description = 'Gaia Bp magnitude'
        tab['gaia_rp_mag'].description = 'Gaia Rp magnitude'
        tab['gaia_variability'].description = 'Gaia variability classification'
        tab['B_mag'].description = 'Johnson B magnitude'
        tab['V_mag'].description = 'Johnson V magnitude'
        tab['H_mag'].description = '2MASS H magnitude'
        tab['J_mag'].description = '2MASS J magnitude'
        tab['K_mag'].description = '2MASS K magnitude'
        tab['u_mag'].description = 'u magnitude'
        tab['g_mag'].description = 'g magnitude'
        tab['r_mag'].description = 'r magnitude'
        tab['i_mag'].description = 'i magnitude'
        tab['z_mag'].description = 'z magnitude'

        # Add some meta-data to the table as well:
        tab.meta['targetid'] = int(dict_tables['target']['targetid'])

        dict_tables[table_name] = tab
        return dict_tables


@lru_cache(maxsize=10)
def get_catalog(target: Union[int, str], radius: Optional[float] = None, output: str = 'table') -> dict:
    """

    Parameters:
        target (int or str):
        radius (float, optional): Radius around target in degrees to return targets for.
        output (str, optional): Desired output format. Choices are 'table', 'dict', 'json'.
            Default='table'.

    Returns:
        dict: Dictionary with three members:
            - 'target': Information about target.
            - 'references': Table with information about reference stars close to target.
            - 'avoid': Table with stars close to target which should be avoided in FOV selection.
    """
    output = OutputFormat(output)
    # make request
    token = get_api_token()
    r = get_request(URLS.catalogs_url, token=token, params={'target': target})
    jsn = r.json()

    # Convert timestamps to actual Time objects:
    jsn['target']['inserted'] = Time(jsn['target']['inserted'], scale='utc')
    if jsn['target']['discovery_date'] is not None:
        jsn['target']['discovery_date'] = Time(jsn['target']['discovery_date'], scale='utc')

    if radius is not None:
        pass  # Not implemented

    if output is not OutputFormat.table:
        return jsn
    return create_catalog_table(jsn)


def get_catalog_missing():
    """
    Get missing catalogs
    Returns: json of missing catalogs
    """
    token = get_api_token()
    r = get_request(URLS.catalogs_missing_url, token=token)
    return r.json()



from astropy.time import Time, TimezoneInfo
from datetime import datetime
from typing import Union


def resolve_date_iso(date: Union[Time, datetime, str]):
    """
    Helper function for resolving varied date input to an iso string.
    Args:
        date: Union[Time, datetime, str] date provided as string, datetime.datetime,
         or astropy.time.Time object
         if string, must be given as UTC!

    Returns: str = iso formatted string representation of the input time
    """
    if isinstance(date, Time):
        return date.utc.iso
    elif isinstance(date, datetime):
        return Time(date.astimezone(TimezoneInfo())).iso  # Defaults to UTC so we don't need pytz.
    elif isinstance(date, str):
        return Time(date).iso  # string assumed to be UTC as per docstring.
    else:
        raise ValueError(f'given date {date} was not a proper astropy.time.Time, datetime.datetime, or UTC string')



"""
For querying photometric filters.
"""

from functools import lru_cache
import astropy.units as u
from tendrils.utils import get_api_token, get_request, URLS


@lru_cache(maxsize=10)
def get_filters() -> dict:
    """
    Get all filters
    Returns: json of filters as python dict

    """
    token = get_api_token()
    r = get_request(URLS.filters_url, token=token)
    jsn = r.json()

    # Add units:
    for f, val in jsn.items():
        if val.get('wavelength_center'):
            val['wavelength_center'] *= u.nm
        if val.get('wavelength_width'):
            val['wavelength_width'] *= u.nm

    return jsn



"""
This subpackage contains utility functions for the Tendrils API.
"""

from .config import *
from .files import get_filehash
from .urls import urls, urls_from_config, get_request, post_request
from .time import resolve_date_iso
from .ztf import download_ztf_photometry, query_ztf_id
from .tns import tns_search, tns_getnames, tns_get_obj, TNSConfigError, load_tns_config

URLS = urls()


"""
For setting up and accessing remote urls. Defaults to FLOWS project ones but you can set up your own.
"""
from dataclasses import dataclass, asdict
import configparser
from tendrils.utils.config import load_config
from typing import Optional
from warnings import warn
import requests


@dataclass
class RemoteUrls:
    base_url: str = 'https://flows.phys.au.dk/api/'
    datafiles_url: str = 'datafiles.php'
    targets_url: str = 'targets.php'
    sites_url: str = 'sites.php'
    set_photometry_status_url: str = 'set_photometry_status.php'
    photometry_url: str = 'download_photometry.php'
    photometry_upload_url: str = 'upload_photometry.php'
    cleanup_photometry_status_url: str = 'cleanup_photometry_status.php'
    catalogs_url: str = 'reference_stars.php'
    catalogs_missing_url: str = 'catalog_missing.php'
    filters_url: str = 'filters.php'
    lightcurves_url: str = 'lightcurve.php'
    targets_post_url: str = 'targets_add.php'

    # def __post_init__(self):
    #    self.urls = [field.name for field in ]
    @staticmethod
    def urls_from_config(config: Optional[configparser.ConfigParser] = None) -> dict:
        """
        Update default remote urls with values found from config. Will warn if ALL values are not found.
        Returns: dict[str:str] representing urls
        These should be under a [URLS] section in config.ini or in another config.
        """
        if config is None:
            config = load_config()

        return dict(config['URL'])

    def update(self, new: Optional[dict] = None) -> None:
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                warn(f'{key} was given with value: {value} but is not one of {asdict(self).keys()}')

    def make_convenient(self) -> None:
        for key, value in asdict(self).items():
            if 'base_url' not in key:
                setattr(self, key, self.base_url + value)


def urls(urls_instance: Optional[RemoteUrls] = None) -> RemoteUrls:
    """
        Convenience function for getting an RemoteUrls instance.
        Warning! Destructively overrides all urls that are not "base_url"
        Returns: RemoteUrls , but with base_url prepended.
        """
    if urls_instance is None:
        urls_instance = RemoteUrls()  # Create default if not given
    urls_instance.make_convenient()
    return urls_instance


def urls_from_config(config: Optional[configparser.ConfigParser] = None) -> RemoteUrls:
    """
    Use config to return RemoteUrls instance with values read from default config or given ConfigParser instance.
    returned RemoteUrls have base_url prepended, similar to urls().
    Args:
        config: Optional[ConfigParser] instance which defaults to None. Uses default config if None.

    Returns: RemoteUrls instance with the urls populated from the given config
    """
    _urls = RemoteUrls()
    new_urls_dict = RemoteUrls.urls_from_config(config)
    _urls.update(new_urls_dict)
    return urls(_urls)


def get_request(url: str, token: str = None, params: Optional[dict] = None,
                headers: Optional[dict] = None) -> requests.Response:
    """
    Make a get request using request with given url, params, and header. Token can be given inplace of header
    to create a default header that uses the token.
    Args:
        url: str = url for get request
        token: Optional[str] = api token. Can also be provided as part of headers dict.
        params: Optional[dict] = dict of params
        headers: Optional[dict] = headers dict. Can also be created from just the token.

    Returns: requests.Response

    """
    params = {} if params is None else params
    headers = {} if headers is None else headers
    if not headers and token is None:
        warn('headers and token was None but one should probably give the authorization token')
    elif token is not None:
        headers['Authorization'] = 'Bearer ' + token

    response = requests.get(url, params=params, headers=headers)
    response.raise_for_status()
    return response


def post_request(url: str, token: str = None, params: Optional[dict] = None, data: Optional[dict] = None,
                 headers: Optional[dict] = None, files: Optional[dict] = None) -> requests.Response:
    """
    Make a post request using request with given url, params, and header. Token can be given inplace of header
    to create a default header that uses the token.
    Args:
        url: str = url for get request
        token: str, optional = api token. Can also be provided as part of headers dict.
        params: dict, optional = dict of params
        data: dict, optional = dict of data
        headers: dict, optional = headers dict. Can also be created from just the token.
        files: dict, optional = files dict.

    Returns: requests.Response
    """
    data = {} if data is None else data
    params = {} if params is None else params
    files = {} if files is None else files
    headers = {} if headers is None else headers
    if not headers and token is None:
        warn('headers and token was None but one should probably give the authorization token')
    elif token is not None:
        headers['Authorization'] = 'Bearer ' + token

    response = requests.post(url, data=data, params=params, files=files, headers=headers)
    response.raise_for_status()
    return response



"""
Set photometry status
"""

import logging
from tendrils.utils import get_api_token, get_request, URLS, load_config


def is_pipeline():
    logger = logging.getLogger(__name__)
    # Get API token from config file:
    config = load_config()
    i_am_pipeline = config.getboolean('api', 'pipeline', fallback=False)
    if not i_am_pipeline:
        logger.debug("Not setting status since user is not pipeline")
        return False
    return True


def set_photometry_status(fileid: int, status: str) -> bool:
    """
    Set photometry status.

    Parameters:
        fileid (int):
        status (str): Choices are 'running', 'error' or 'done'.

    .. codeauthor:: Rasmus Handberg <rasmush@phys.au.dk>
    """
    # Validate the input:
    if status not in ('running', 'error', 'abort', 'ingest', 'done'):
        raise ValueError('Invalid status')

    if not is_pipeline():
        return False

    # Get API token from config file:
    token = get_api_token()

    # Send HTTP request to FLOWS server:
    r = get_request(URLS.set_photometry_status_url, token=token, params={'fileid': fileid, 'status': status})
    result = r.text.strip()
    if result != 'OK':
        raise RuntimeError(result)

    return True


def cleanup_photometry_status() -> bool:
    """
    Perform a cleanup of the photometry status indicator.

    This will change all processes still marked as "running"
    to "abort" if they have been running for more than a day.
    """
    # Exit if I am not the pipeline
    if not is_pipeline():
        return False

    # Get API token from config file:
    token = get_api_token()

    # Send HTTP request to FLOWS server:
    r = get_request(URLS.cleanup_photometry_status_url, token=token)
    result = r.text.strip()
    if result != 'OK':
        raise RuntimeError(result)

    return True



"""
Functions for accessing flows datafiles from a remote server
"""

from datetime import datetime
from functools import lru_cache
from typing import Union, Optional
from tendrils.utils import get_api_token, get_request, URLS

@lru_cache(maxsize=10)
def get_datafile(fileid: Union[int, str]):
    """
    Get single datafile
    Args:
        fileid: fileid to query

    Returns: json of datafile
    """
    token = get_api_token()
    r = get_request(URLS.datafiles_url, token=token, params={'fileid': fileid})
    jsn = r.json()

    # Parse some of the fields to Python objects:
    jsn['inserted'] = datetime.strptime(jsn['inserted'], '%Y-%m-%d %H:%M:%S.%f')
    jsn['lastmodified'] = datetime.strptime(jsn['lastmodified'], '%Y-%m-%d %H:%M:%S.%f')

    return jsn


def get_datafiles(targetid: Optional[int] = None,
                  filt: str = 'missing',
                  minversion: Optional[str] = None) -> list[int]:
    """
    Get list of data file IDs to be processed.

    Parameters:
        targetid (int, optional): Target ID to process.
        filt (str, optional): Filter the returned list:
            - ``missing``: Return only data files that have not yet been processed.
            - ``'all'``: Return all data files.
        minversion (str, optional): Special filter matching files not processed at least with
            the specified version (defined internally in API for now).

    Returns:
        list: List of data files the can be processed.
    """

    # Validate input:
    if filt not in ('missing', 'all', 'error'):
        raise ValueError("Invalid filter specified: '%s'" % filt)

    token = get_api_token()

    params = {}
    if targetid is not None:
        params['targetid'] = targetid
    if minversion is not None:
        params['minversion'] = minversion
    params['filter'] = filt

    r = get_request(URLS.datafiles_url, token=token, params=params)
    return r.json()



"""
Get information about targets in Flows.
Add target to Flows.
"""

import re
from datetime import datetime
from functools import lru_cache
from typing import Union, Optional
from astropy.time import Time
from astropy.coordinates import SkyCoord
from tendrils.utils import get_api_token, get_request, URLS, resolve_date_iso, post_request

@lru_cache(maxsize=10)
def get_target(target: Union[int, str]) -> dict:
    """
    Get target as json
    Args:
        target: Optional[int, str] targetid

    Returns: json of target info

    """
    token = get_api_token()
    r = get_request(URLS.targets_url, token=token, params={'target': target})
    jsn = r.json()

    # Parse some of the fields to Python objects:
    jsn['inserted'] = datetime.strptime(jsn['inserted'], '%Y-%m-%d %H:%M:%S.%f')
    if jsn['discovery_date']:
        jsn['discovery_date'] = Time(jsn['discovery_date'], format='iso', scale='utc')

    return jsn


@lru_cache(maxsize=1)
def get_targets() -> dict:
    """
    Get json list of all targets and some basic info about them
    Returns: json list of all targets

    """
    token = get_api_token()
    r = get_request(URLS.targets_url, token=token)
    jsn = r.json()

    # Parse some of the fields to Python objects:
    for tgt in jsn:
        tgt['inserted'] = datetime.strptime(tgt['inserted'], '%Y-%m-%d %H:%M:%S.%f')
        if tgt['discovery_date']:
            tgt['discovery_date'] = Time(tgt['discovery_date'], format='iso', scale='utc')

    return jsn


def add_target(name: str,
               coord: SkyCoord,
               redshift: Optional[float] = None,
               redshift_error: Optional[float] = None,
               discovery_date: Union[Optional[Time], Optional[datetime], Optional[str]] = None,
               discovery_mag: Union[Optional[float], Optional[int]] = None,
               host_galaxy: Optional[str] = None,
               ztf: Optional[str] = None,
               sntype: Optional[str] = None,
               status: str = 'candidate',
               project: str = 'flows') -> int:
    """
    Add new candidate or target.

    Coordinates are specified using an Astropy SkyCoord object, which can be
    created in the following way:

    coord = SkyCoord(ra=19.1, dec=89.00001, unit='deg', frame='icrs')

    The easiest way is to specify ``discovery_date`` as an Astropy Time object:

    discovery_date = Time('2020-01-02 00:00:00', format='iso', scale='utc')

    Alternatively, you can also specify it as a :class:`datetime.datetime` object,
    but some care has to be taken with specifying the correct timezone:

    discovery_date = datetime.strptime('2020-01-02 00:00:00', '%Y-%m-%d %H:%M:%S%f')
    discovery_date = pytz.timezone('America/New_York').localize(discovery_date)

    Lastly, it can be given as a simple date-string of the following form,
    but here the data has to be given in UTC:

    discovery_date = '2020-01-02 23:56:02.123'

    Parameters:
        name (str): Name of target. Must be of the form "YYYYxyz", where YYYY is the year,
            and xyz are letters.
        coord (:class:ʼastropy.coordinates.SkyCoordʼ): Sky coordinates of target.
        redshift (float, optional): Redshift.
        redshift_error (float, optional): Uncertainty on redshift.
        discovery_date (:class:`astropy.time.Time`, :class:`datetime.datetime` or str, optional):
        discovery_mag (float, int, optional): Magnitude at time of discovery.
        host_galaxy (str, optional): Host galaxy name.
        sntype (str, optional): Supernovae type (e.g. Ia, Ib, II).
        ztf (str, optional): ZTF identifier.
        status (str, optional):
        project (str, optional):

    Returns:
        int: New target identifier in Flows system.
    """
    # Check and convert input:
    if not re.match(r'^[12]\d{3}([A-Z]|[a-z]{2,4})$', name.strip()):
        raise ValueError("Invalid target name.")

    if redshift is None and redshift_error is not None:
        raise ValueError("Redshift error specified without redshift value")

    discovery_date = resolve_date_iso(discovery_date)  # Get as UTC iso string, raises if invalid
    discovery_mag = float(discovery_mag) if discovery_mag is not None else discovery_mag

    if status not in ('candidate', 'target'):
        raise ValueError("Invalid target status.")

    token = get_api_token()

    # Gather parameters to be sent to API:
    params = {'targetid': 0, 'target_name': name.strip(), 'ra': coord.icrs.ra.deg, 'decl': coord.icrs.dec.deg,
              'redshift': redshift, 'redshift_error': redshift_error, 'discovery_date': discovery_date,
              'discovery_mag': discovery_mag, 'host_galaxy': host_galaxy, 'project': project, 'ztf_id': ztf,
              'target_status': status, 'sntype': sntype}

    # Post the request to the API:
    r = post_request(URLS.targets_post_url, token=token, data=params)
    jsn = r.json()

    # Check for errors:
    if jsn['errors'] is not None:
        raise RuntimeError(f"Adding target '{name}' resulted in an error: {jsn['errors']}")

    return int(jsn['targetid'])



"""
Query ZTF target information using ALeRCE API.
https://alerceapi.readthedocs.io/
@TODO: REFACTOR TO FIT API! Make Class.
"""
from __future__ import annotations
import numpy as np
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.table import Table
from astropy.time import Time
import datetime
import requests
from tendrils import api
from typing import Optional


def query_ztf_id(coo_centre: SkyCoord, radius: u.Quantity = 3 * u.arcsec, discovery_date: Optional[Time] = None) \
        -> Optional[str]:
    """
    Query ALeRCE ZTF api to lookup ZTF identifier.

    In case multiple identifiers are found within the search cone, the one
    closest to the centre is returned.

    Parameters:
        coo_centre (:class:`astropy.coordinates.SkyCoord`): Coordinates of centre of search cone.
        radius (Angle, optional): Search radius. Default 3 arcsec.
        discovery_date (:class:`astropy.time.Time`, optional): Discovery date of target to
            match against ZTF. The date is compared to the ZTF first timestamp and ZTF targets
            are rejected if they are not within 15 days prior to the discovery date
            and 90 days after.

    Returns:
        str: ZTF identifier.
    """

    if isinstance(radius, (float, int)):
        radius *= u.deg

    # Make json query for Alerce query API
    query = {'ra': coo_centre.ra.deg, 'dec': coo_centre.dec.deg, 'radius': Angle(radius).arcsec, 'page_size': 20,
             'count': True}

    # Run http POST json query to alerce following their API
    res = requests.get('https://api.alerce.online/ztf/v1/objects', params=query)
    res.raise_for_status()
    jsn = res.json()

    # If nothing was found, return None:
    if jsn['total'] == 0:
        return None

    # Start by removing anything marked as likely stellar-like source:
    results = jsn['items']
    results = [itm for itm in results if not itm['stellar']]
    if not results:
        return None

    # Constrain on the discovery date if it is provided:
    if discovery_date is not None:
        # Extract the time of the first ZTF timestamp and compare it with
        # the discovery time:
        firstmjd = Time([itm['firstmjd'] for itm in results], format='mjd', scale='utc')
        tdelta = firstmjd.utc.mjd - discovery_date.utc.mjd

        # Only keep results that are within the margins:
        results = [itm for k, itm in enumerate(results) if -15 <= tdelta[k] <= 90]
        if not results:
            return None

    # Find target closest to the centre:
    coords = SkyCoord(ra=[itm['meanra'] for itm in results], dec=[itm['meandec'] for itm in results], unit='deg',
                      frame='icrs')

    indx = np.argmin(coords.separation(coo_centre))

    return results[indx]['oid']


def download_ztf_photometry(targetid: int) -> Table:
    """
    Download ZTF photometry from ALERCE API.

    Parameters:
        targetid (int): Target identifier.

    Returns:
        :class:`astropy.table.Table`: ZTF photometry table.
    """

    # Get target info from Flows API:
    tgt = api.get_target(targetid)
    oid = tgt['ztf_id']
    target_name = tgt['target_name']
    if oid is None:
        return None

    # Query ALERCE for detections of object based on oid
    res = requests.get(f'https://api.alerce.online/ztf/v1/objects/{oid:s}/detections')
    res.raise_for_status()
    jsn = res.json()

    # Create Astropy table, cut out the needed columns
    # and rename columns to something better for what we are doing:
    tab = Table(data=jsn)
    tab = tab[['fid', 'mjd', 'magpsf', 'sigmapsf']]
    tab.rename_column('fid', 'photfilter')
    tab.rename_column('mjd', 'time')
    tab.rename_column('magpsf', 'mag')
    tab.rename_column('sigmapsf', 'mag_err')

    # Remove bad values of time and magnitude:
    tab['time'] = np.asarray(tab['time'], dtype='float64')
    tab['mag'] = np.asarray(tab['mag'], dtype='float64')
    tab['mag_err'] = np.asarray(tab['mag_err'], dtype='float64')
    indx = np.isfinite(tab['time']) & np.isfinite(tab['mag']) & np.isfinite(tab['mag_err'])
    tab = tab[indx]

    # Replace photometric filter numbers with keywords used in Flows:
    photfilter_dict = {1: 'gp', 2: 'rp', 3: 'ip'}
    tab['photfilter'] = [photfilter_dict[fid] for fid in tab['photfilter']]

    # Sort the table on photfilter and time:
    tab.sort(['photfilter', 'time'])

    # Add meta information to table header:
    tab.meta['target_name'] = target_name
    tab.meta['targetid'] = targetid
    tab.meta['ztf_id'] = oid
    tab.meta['last_updated'] = datetime.datetime.now(tz=datetime.timezone.utc).isoformat()

    return tab



import setuptools
import sys
# Workaround for dynamic version reading..
sys.path[0:0] = ['tendrils']
from version import version as __version__  # noqa: E402
with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='tendrils',
    version=__version__,
    author="Emir Karamehmetoglu, Rasmus Handberg",
    author_email="emirkmo@github.com",
    description="Tendrils: API for accessing flows pipeline and data products.",
    long_description=long_description,
    packages=setuptools.find_packages(),
    license="LICENSE.rst",
    url="https://github.com/SNFLOWS/tendrils",
    classifiers=["Programming Language :: Python :: 3",
                 "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
                 "Operating System :: OS Independent",
                 "Topic :: Scientific/Engineering :: Astronomy"],
    python_requires='>=3.10',
    install_requires=['numpy',
                      'astropy',
                      'requests',
                      'tqdm',
                      ],
    package_data={"tendrils": ["utils/*.ini"]}  # include any .ini found in utils.
)



"""
API subpackage. Also contains base functions.
"""
# flake8: noqa

from .catalogs import get_catalog, get_catalog_missing
from .datafiles import get_datafile, get_datafiles
from .filters import get_filters
from .lightcurves import get_lightcurve
from .photometry_api import get_photometry, upload_photometry
from .set_photometry_status import set_photometry_status, cleanup_photometry_status
from .sites import get_site, get_all_sites
from .targets import get_targets, get_target, add_target



"""
Utility functions for dealing with files.
"""
import hashlib


def get_filehash(filename: str):
    """
    Calculate SHA1-hash of file. Raise exception if length is invalid.
    Args:
        filename: str = filename
    Returns: sha1sum file hash
    """
    buf = 65536
    s = hashlib.sha1()
    with open(filename, 'rb') as fid:
        while True:
            data = fid.read(buf)
            if not data:
                break
            s.update(data)

    sha1sum = s.hexdigest().lower()
    if len(sha1sum) != 40:
        raise Exception("Invalid file hash")
    return sha1sum
